import numpy as np
import matplotlib.pyplot as plt
import inspect # Used for storing the input
from .element import Element
from .equation import HeadEquation, HeadEquationNores, \
                      MscreenEquation, MscreenDitchEquation
from . import besselnumba

class LineSinkBase(Element):
    '''LineSink Base Class. All LineSink elements are derived from this class'''
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, tsandbc=[(0, 1)], \
                 res=0, wh='H', layers=0, type='', name='LineSinkBase', \
                 label=None, addtomodel=True):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         tsandbc=tsandbc, type=type, name=name, label=label)
        self.nparam = len(self.layers)
        self.x1 = float(x1)
        self.y1 = float(y1)
        self.x2 = float(x2)
        self.y2 = float(y2)
        self.res = np.atleast_1d(res).astype(float)
        self.wh = wh
        if addtomodel: self.model.addelement(self)

    def __repr__(self):
        return self.name + ' from ' + str((self.x1, self.y1)) + \
                           ' to ' + str((self.x2,self.y2))

    def initialize(self):
        self.xc = np.array([0.5 * (self.x1 + self.x2)])
        self.yc = np.array([0.5 * (self.y1 + self.y2)])
        self.ncp = 1
        self.z1 = self.x1 + 1j*self.y1; self.z2 = self.x2 + 1j*self.y2
        self.L = np.abs(self.z1-self.z2)
        self.order = 0 # This is for univform discharge only
        self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        self.setbc()
        coef = self.aq.coef[self.layers, :]
        self.setflowcoef()
        # self.term shape (self.nparam,self.aq.naq,self.model.npval)
        self.term = self.flowcoef * coef  
        self.term2 = self.term.reshape(
            self.nparam, self.aq.naq, self.model.nint, self.model.npint)
        self.dischargeinf = self.flowcoef * coef
        self.dischargeinflayers = np.sum(
            self.dischargeinf * self.aq.eigvec[self.layers, :, :], 1)
        if type(self.wh) is str:
            if self.wh == 'H':
                self.wh = self.aq.Haq[self.layers]
            elif self.wh == '2H':
                self.wh = 2.0 * self.aq.Haq[self.layers]
        else:
            self.wh = np.atleast_1d(self.wh) * np.ones(self.nlayers)
        # Q = (h - hls) / resfach
        self.resfach = self.res / (self.wh * self.L)  
        # Q = (Phi - Phils) / resfacp
        self.resfacp = self.resfach * self.aq.T[self.layers]  

    def setflowcoef(self):
        '''Separate function so that this can be overloaded for other types'''
        self.flowcoef = 1.0 / self.model.p  # Step function

    def potinf(self, x, y, aq=None):
        '''Can be called with only one x,y value'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros(
            (self.nparam, aq.naq, self.model.nint, self.model.npint), 'D')
        if aq == self.aq:
            pot = np.zeros(self.model.npint, 'D')
            for i in range(self.aq.naq):
                for j in range(self.model.nint):
                    pot[:] = besselnumba.bessellsuniv(x, y, 
                        self.z1, self.z2, self.aq.lab2[i, j, :], self.rzero)
                    # Divide by L as the parameter is total discharge
                    rv[:,i,j,:] = self.term2[:, i, j, :] * pot / self.L  
        rv.shape = (self.nparam, aq.naq, self.model.npval)
        return rv

    def disvecinf(self,x,y,aq=None):
        '''Can be called with only one x,y value'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rvx = np.zeros(
            (self.nparam, aq.naq, self.model.nint, self.model.npint), 'D')
        rvy = np.zeros(
            (self.nparam, aq.naq, self.model.nint, self.model.npint), 'D')
        if aq == self.aq:
            qxqy = np.zeros((2,self.model.npint), 'D')
            for i in range(self.aq.naq):
                for j in range(self.model.nint):
                    if besselnumba.isinside(self.z1, self.z2, x + y * 1j, 
                                       self.rzero * self.aq.lababs[i, j]):
                        qxqy[:,:] = besselnumba.bessellsqxqyv2(x, y, 
                            self.z1, self.z2, self.aq.lab2[i, j, :], 
                            self.order, 
                            self.rzero * self.aq.lababs[i, j]) / self.L
                    rvx[:,i,j,:] = self.term2[:,i,j,:] * qxqy[0]
                    rvy[:,i,j,:] = self.term2[:,i,j,:] * qxqy[1]
        rvx.shape = (self.nparam, aq.naq, self.model.npval)
        rvy.shape = (self.nparam, aq.naq, self.model.npval)
        return rvx, rvy

    def headinside(self,t):
        """The head inside the line-sink
        
        Parameters
        ----------
        t : array or float
            time(s) for whih head is computed
        
        Returns
        -------
        array (length number of layers)
            Head inside the line-sink for each layer that 
            the line-sink is screened in
            
        """
        
        return self.model.head(self.xc[0],self.yc[0],t)[self.layers] - \
                   self.resfach[:, np.newaxis] * self.discharge(t)

    def plot(self):
        plt.plot([self.x1, self.x2], [self.y1, self.y2], 'k')
        
class LineSink(LineSinkBase):
    '''LineSink with non-zero and potentially variable discharge through time
    really only used for testing'''
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, tsandQ=[(0, 1)],
                 res=0, wh='H', layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        LineSinkBase.__init__(self, model, x1=x1, y1=y1, x2=x2, y2=y2, 
                              tsandbc=tsandQ, res=res, wh=wh, layers=layers, 
                              type='g', name='LineSink', label=label,
                              addtomodel=addtomodel)

    
#class ZeroHeadLineSink(LineSinkBase,HeadEquation):
    #'''HeadLineSink that remains zero and constant through time'''
    #def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, res=0.0, wh='H', \
    #             layers=0, label=None, addtomodel=True):
    #    self.storeinput(inspect.currentframe())
    #    LineSinkBase.__init__(self, model, x1=x1, y1=y1, x2=x2, y2=y2, \
    #                          tsandbc=[(0, 0)], res=res, wh=wh, layers=layers,\
    #                          type='z', name='ZeroHeadLineSink', label=label, \
    #                          addtomodel=addtomodel)
    #    self.nunknowns = self.nparam
    #    
    #def initialize(self):
    #    LineSinkBase.initialize(self)
    #    self.parameters = np.zeros((self.model.ngvbc, self.nparam, 
    #                                self.model.npval), 'D')
        
class HeadLineSink(LineSinkBase, HeadEquation):
    """
    Create a head-specified line-sink
    which may optionally have a width and resistance
    Inflow per unit length of line-sink is computed as
    
    .. math::
        \sigma = w(h_{aq} - h_{ls})/c
    
    where :math:`c` is the resistance of the bottom of the line-sink,
    :math:`w` is the width over which water enters the line-sink,
    :math:`h_{aq}` is the head in the aquifer at the center of the line-sink,
    :math:`h_{ls}` is the specified head inside the line-sink
    Note that all that matters is the conductance term :math:`w/c` but
    both are specified separately
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    x1 : scalar
        x-coordinate of fist point of line-sink
    y1 : scalar
        y-coordinate of fist point of line-sink
    x2 : scalar
        x-coordinate of second point of line-sink
    y2 : scalar
        y-coordinate of second point of line-sink
    tsandh : list or 2D array of (time, head) values or string
        if list or 2D array: pairs of time and head after that time
        if 'fixed': head is fixed (no change in head) during entire simulation
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer 
        layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when 
        flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the 
        aquifer layer
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.HeadLineSinkString`
    
    """
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, tsandh=[(0, 1)],
                 res=0, wh='H', layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        if tsandh == 'fixed':
            tsandh = [(0, 0)]
            etype = 'z'
        else:
            etype = 'v'
        LineSinkBase.__init__(self, model, x1=x1, y1=y1, x2=x2, y2=y2,
                              tsandbc=tsandh, res=res, wh=wh, layers=layers,
                              type=etype, name='HeadLineSink', label=label,
                              addtomodel=addtomodel)
        self.nunknowns = self.nparam

    def initialize(self):
        LineSinkBase.initialize(self)
        self.parameters = np.zeros((self.model.ngvbc, self.nparam, 
                                    self.model.npval), 'D')
        # Needed in solving, solve for a unit head
        self.pc = self.aq.T[self.layers] 
        
class LineSinkStringBase(Element):
    def __init__(self, model, tsandbc=[(0, 1)], layers=0, type='',
                 name='LineSinkStringBase', label=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         tsandbc=tsandbc, type=type, name=name, label=label)
        self.lslist = []

    def __repr__(self):
        return self.name + ' with nodes ' + str(zip(self.x,self.y))

    def initialize(self):
        self.ncp = self.nls
        self.nparam = self.nlayers * self.nls
        self.nunknowns = self.nparam
        self.xls,self.yls = np.empty((self.nls,2)), np.empty((self.nls,2))
        for i,ls in enumerate(self.lslist):
            ls.initialize()
            self.xls[i,:] = [ls.x1,ls.x2]
            self.yls[i,:] = [ls.y1,ls.y2]
        # Only used for layout when it is a continuous string
        self.xlslayout = np.hstack((self.xls[:,0],self.xls[-1,1])) 
        self.ylslayout = np.hstack((self.yls[:,0],self.yls[-1,1]))
        self.aq = self.model.aq.find_aquifer_data(
            self.lslist[0].xc, self.lslist[0].yc)
        self.parameters = np.zeros((self.model.ngvbc, self.nparam, 
                                    self.model.npval), 'D')
        self.setbc()
        # As parameters are only stored for the element not the list
        # we need to combine the following
        self.resfach = []; self.resfacp = []
        for ls in self.lslist:
            ls.initialize()
            self.resfach.extend(ls.resfach.tolist())  # Needed in solving
            self.resfacp.extend(ls.resfacp.tolist())  # Needed in solving
        self.resfach = np.array(self.resfach)
        self.resfacp = np.array(self.resfacp)
        self.dischargeinf = np.zeros((self.nparam, self.aq.naq, 
                                      self.model.npval), 'D')
        self.dischargeinflayers = np.zeros((self.nparam, self.model.npval), 'D')
        self.xc, self.yc = np.zeros(self.nls), np.zeros(self.nls)
        for i in range(self.nls):
            self.dischargeinf[
                i * self.nlayers: (i + 1) * self.nlayers, :] = \
                self.lslist[i].dischargeinf[:]
            self.dischargeinflayers[
                i * self.nlayers: (i + 1) * self.nlayers, :] = \
                self.lslist[i].dischargeinflayers
            self.xc[i] = self.lslist[i].xc
            self.yc[i] = self.lslist[i].yc

    def potinf(self, x, y,aq=None):
        '''Returns array (nunknowns, Nperiods)'''
        if aq is None: 
            aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        for i in range(self.nls):
            rv[i * self.nlayers: (i + 1) * self.nlayers, :] = \
                self.lslist[i].potinf(x, y, aq)
        return rv

    def disvecinf(self,x,y,aq=None):
        '''Returns array (nunknowns,Nperiods)'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rvx = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        rvy = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        for i in range(self.nls):
            qx, qy = self.lslist[i].disvecinf(x, y, aq)
            rvx[i * self.nlayers: (i + 1) * self.nlayers, :] = qx
            rvy[i * self.nlayers: (i + 1) * self.nlayers, :] = qy
        return rvx,rvy

    def headinside(self, t, derivative=0):
        """The head inside the line-sink string
        
        Parameters
        ----------
        t : array or float
            time(s) for whih head is computed
        
        Returns
        -------
        array size nline-sinks, nlayers, ntimes
            Head inside the line-sink for each line-sink, each layer that
            the line-sink is screened in, and each time
            
        """
        
        rv = np.zeros((self.nls, self.nlayers, np.size(t)))
        Q = self.discharge_list(t, derivative=derivative)
        for i in range(self.nls):
            rv[i, :, :] = self.model.head(self.xc[i], self.yc[i], t,
                derivative=derivative)[self.layers] - \
                self.resfach[i * self.nlayers: (i + 1) * self.nlayers,
                             np.newaxis] * Q[i]
        return rv
    
    def plot(self):
        plt.plot(self.xlslayout, self.ylslayout, 'k')

    def run_after_solve(self):
        for i in range(self.nls):
            self.lslist[i].parameters[:] = \
                self.parameters[:, i * self.nlayers:(i + 1) * self.nlayers, :]

    def discharge_list(self,t,derivative=0):
        """The discharge of each line-sink in the string
        
        Parameters
        ----------
        t : array or float
            time(s) for whih discharge is computed
        
        Returns
        -------
        array size nline-sinks, nlayers, ntimes
            Discharge for each line-sink, each layer that
            the line-sink is screened in, and each time
            
        """
        rv = np.zeros((self.nls, self.nlayers, np.size(t)))
        for i in range(self.nls):
            rv[i,:,:] = self.lslist[i].discharge(t,derivative=derivative)
        return rv
            
class HeadLineSinkString(LineSinkStringBase, HeadEquation):
    """
    Create string of head-specified line-sinks
    which may optionally have a width and resistance
    Inflow per unit length of line-sink is computed as
    
    .. math::
        \sigma = w(h_{aq} - h_{ls})/c
    
    where :math:`c` is the resistance of the bottom of the line-sink,
    :math:`w` is the width over which water enters the line-sink,
    :math:`h_{aq}` is the head in the aquifer at the center of the line-sink,
    :math:`h_{ls}` is the specified head inside the line-sink
    Note that all that matters is the conductance term :math:`w/c` but
    both are specified separately
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        line-sinks in string
    tsandh : list or 2D array of (time, head) values or string
        if list or 2D array: pairs of time and head after that time
        if 'fixed': head is fixed (no change in head) during entire simulation
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer 
        layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when 
        flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the 
        aquifer layer
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.HeadLineSink`
    
    """
    def __init__(self, model, xy=[(-1, 0), (1, 0)], tsandh=[(0, 1)],
                 res=0, wh='H', layers=0, label=None):
        if tsandh == 'fixed':
            tsandh = [(0, 0)]
            etype = 'z'
        else:
            etype = 'v'
        LineSinkStringBase.__init__(self, model, tsandbc=tsandh, layers=layers,
                                    type=etype, name='HeadLineSinkString', 
                                    label=label)
        xy = np.atleast_2d(xy).astype('d')
        self.x = xy[:, 0]
        self.y = xy[:, 1]
        self.nls = len(self.x) - 1
        self.tsandh = tsandh
        self.res = np.atleast_1d(float(res))
        self.wh = wh
        self.model.addelement(self)

    def initialize(self):
        self.lslist = []
        for i in range(self.nls):
            self.lslist.append(HeadLineSink(
                self.model, x1=self.x[i], y1=self.y[i], x2=self.x[i + 1], 
                y2=self.y[i + 1], tsandh=self.tsandh, res=self.res, wh=self.wh,               
                layers=self.layers, label=None, addtomodel=False))
        LineSinkStringBase.initialize(self)
        self.pc = np.zeros(self.nls * self.nlayers)
        for i in range(self.nls):
            self.pc[i * self.nlayers:(i + 1) * self.nlayers] = self.lslist[i].pc
            
class MscreenLineSink(LineSinkBase,MscreenEquation):
    '''MscreenLineSink that varies through time. 
    Must be screened in multiple layers but heads are same 
    in all screened layers'''
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, tsandQ=[(0.0, 1.0)],
                 res=0.0, wh='H', layers=[0, 1], vres=0.0, wv=1.0, label=None,
                 addtomodel=True):
        #assert len(layers) > 1, "number of layers must be at least 2"
        self.storeinput(inspect.currentframe())
        LineSinkBase.__init__(self, model, x1=x1, y1=y1, x2=x2, y2=y2, 
                              tsandbc=tsandQ, res=res, wh=wh, layers=layers, 
                              type='v', name='MscreenLineSink', label=label, 
                              addtomodel=addtomodel)
        self.nunknowns = self.nparam
        # Vertical resistance inside line-sink
        self.vres = np.atleast_1d(vres)  
        self.wv = wv
        if len(self.vres) == 1: 
            self.vres = self.vres[0] * np.ones(self.nlayers - 1)
    def initialize(self):
        LineSinkBase.initialize(self)
        self.parameters = np.zeros((self.model.ngvbc, self.nparam, 
                                    self.model.npval), 'D')
        # Qv = (hn - hn-1) / vresfac[n - 1]
        self.vresfac = self.vres / (self.wv * self.L)  
            
class LineSinkDitchString(LineSinkStringBase, MscreenDitchEquation):
    """
    Create ditch consisting of a string of line-sink.
    The total discharge for the string is specified and divided over the
    line-sinks such that the head at the center inside each line-sink is
    equal. A width and resistance may optionally be specified.
    Inflow per unit length of line-sink is computed as
    
    .. math::
        \sigma = w(h_{aq} - h_{ls})/c
    
    where :math:`c` is the resistance of the bottom of the line-sink,
    :math:`w` is the width over which water enters the line-sink,
    :math:`h_{aq}` is the head in the aquifer at the center of the line-sink,
    :math:`h_{ls}` is the specified head inside the line-sink
    Note that all that matters is the conductance term :math:`w/c` but
    both are specified separately
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        line-sinks in string
    tsandQ : list or 2D array of (time, discharge) values
        if list or 2D array: pairs of time and discharge after that time
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer 
        layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when 
        flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the 
        aquifer layer
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element
        
    """    
    def __init__(self, model, xy=[(-1, 0), (1, 0)], tsandQ=[(0, 1)], res=0,
                 wh='H', layers=0, Astorage=None, label=None):
        self.storeinput(inspect.currentframe())
        LineSinkStringBase.__init__(self, model, tsandbc=tsandQ, layers=layers,
                                    type='v', name='LineSinkDitchString',
                                    label=label)
        xy = np.atleast_2d(xy).astype('d')
        self.x,self.y = xy[:, 0], xy[:, 1]
        self.nls = len(self.x) - 1
        for i in range(self.nls):
            self.lslist.append(
                MscreenLineSink(model, x1=self.x[i], y1=self.y[i], 
                                x2=self.x[i + 1], y2=self.y[i + 1], 
                                tsandQ=tsandQ, res=res, wh=wh, layers=layers, 
                                label=None, addtomodel=False))
        self.Astorage = Astorage
        self.model.addelement(self)
    def initialize(self):
        LineSinkStringBase.initialize(self)
        # set vresfac to zero, as I don't quite know what it would mean if 
        # it is not zero
        self.vresfac = np.zeros_like(self.resfach) 
        
class LineSinkDitchString2(LineSinkStringBase, MscreenDitchEquation):
    """
    Create ditch consisting of a string of line-sink.
    The total discharge for the string is specified and divided over the
    line-sinks such that the head at the center inside each line-sink is
    equal. A width and resistance may optionally be specified.
    Inflow per unit length of line-sink is computed as
    
    .. math::
        \sigma = w(h_{aq} - h_{ls})/c
    
    where :math:`c` is the resistance of the bottom of the line-sink,
    :math:`w` is the width over which water enters the line-sink,
    :math:`h_{aq}` is the head in the aquifer at the center of the line-sink,
    :math:`h_{ls}` is the specified head inside the line-sink
    Note that all that matters is the conductance term :math:`w/c` but
    both are specified separately
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        line-sinks in string
    tsandQ : list or 2D array of (time, discharge) values
        if list or 2D array: pairs of time and discharge after that time
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer 
        layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when 
        flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the 
        aquifer layer
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element
        
    """    
    def __init__(self, model, xy=[(-1, 0), (1, 0)], tsandQ=[(0, 1)], res=0,
                 wh='H', layers=0, Astorage=None, label=None):
        self.storeinput(inspect.currentframe())
        LineSinkStringBase.__init__(self, model, tsandbc=tsandQ, layers=layers,
                                    type='v', name='LineSinkDitchString',
                                    label=label)
        xy = np.atleast_2d(xy).astype('d')
        self.x,self.y = xy[:, 0], xy[:, 1]
        self.nls = len(self.x) - 1
        for i in range(self.nls):
            self.lslist.append(
                MscreenLineSink(model, x1=self.x[i], y1=self.y[i], 
                                x2=self.x[i + 1], y2=self.y[i + 1], 
                                tsandQ=tsandQ, res=res, wh=wh, layers=layers, 
                                label=None, addtomodel=False))
        self.Astorage = Astorage
        self.model.addelement(self)
    def initialize(self):
        LineSinkStringBase.initialize(self)
        # set vresfac to zero, as I don't quite know what it would mean if 
        # it is not zero
        self.vresfac = np.zeros_like(self.resfach) 

class LineSinkHoBase(Element):
    '''Higher Order LineSink Base Class. All Higher Order Line Sink elements 
    are derived from this class'''
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, tsandbc=[(0.0,1.0)],
                 res=0.0, wh='H', order=0, layers=0, type='',
                 name='LineSinkBase', label=None, addtomodel=True):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         tsandbc=tsandbc, type=type, name=name, label=label)
        self.order = order
        self.nparam = (self.order + 1) * len(self.layers)
        self.x1 = float(x1)
        self.y1 = float(y1)
        self.x2 = float(x2)
        self.y2 = float(y2)
        self.res = res
        self.wh = wh
        if addtomodel: self.model.addelement(self)

    def __repr__(self):
        return self.name + ' from ' + str((self.x1, self.y1)) + \
               ' to ' + str((self.x2, self.y2))

    def initialize(self):
        self.ncp = self.order + 1
        self.z1 = self.x1 + 1j * self.y1
        self.z2 = self.x2 + 1j * self.y2
        self.L = np.abs(self.z1 - self.z2)
        #
        thetacp = np.arange(np.pi, 0, -np.pi / self.ncp) - \
            0.5 * np.pi / self.ncp
        Zcp = np.zeros(self.ncp, 'D')
        Zcp.real = np.cos(thetacp)
        # control point just on positive site (this is handy later on)
        Zcp.imag = 1e-6  
        zcp = Zcp * (self.z2 - self.z1) / 2 + 0.5 * (self.z1 + self.z2)
        self.xc = zcp.real
        self.yc = zcp.imag
        #
        self.aq = self.model.aq.find_aquifer_data(self.xc[0], self.yc[0])
        self.setbc()
        coef = self.aq.coef[self.layers, :]
        self.setflowcoef()
        # shape of term (self.nlayers, self.aq.naq, self.model.npval)
        self.term = self.flowcoef * coef 
        self.term2 = self.term.reshape(self.nlayers, self.aq.naq, 
                                       self.model.nint, self.model.npint)
        self.dischargeinf = self.flowcoef * coef
        self.dischargeinflayers = np.sum(self.dischargeinf * 
                                         self.aq.eigvec[self.layers, :, :], 1)
        if self.wh == 'H':
            self.wh = self.aq.Haq[self.layers]
        elif self.wh == '2H':
            self.wh = 2.0 * self.aq.Haq[self.layers]
        else:
            self.wh = np.atleast_1d(self.wh) * np.ones(self.nlayers)
        # Q = (h - hls) / resfach
        self.resfach = self.res / (self.wh * self.L)  
        # Q = (Phi - Phils) / resfacp
        self.resfacp = self.resfach * self.aq.T[self.layers]  

    def setflowcoef(self):
        '''Separate function so that this can be overloaded for other types'''
        self.flowcoef = 1 / self.model.p  # Step function

    def potinf(self,x,y,aq=None):
        '''Can be called with only one x,y value'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq, self.model.nint, 
                       self.model.npint), 'D')
        if aq == self.aq:
            pot = np.zeros((self.order + 1, self.model.npint), 'D')
            for i in range(self.aq.naq):
                for j in range(self.model.nint):
                    if besselnumba.isinside(self.z1, self.z2, x + y * 1j, 
                            self.rzero * self.aq.lababs[i, j]):
                        pot[:,:] = besselnumba.bessellsv2(x, y, 
                            self.z1, self.z2, self.aq.lab2[i, j, :],
                            self.order, self.rzero * self.aq.lababs[i, j]
                            ) / self.L  
                    for k in range(self.nlayers):
                        rv[k::self.nlayers, i, j, :] = \
                            self.term2[k, i, j, :] * pot
                        
        rv.shape = (self.nparam, aq.naq, self.model.npval)
        return rv

    def disvecinf(self, x, y, aq=None):
        '''Can be called with only one x,y value'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rvx = np.zeros((self.nparam, aq.naq, self.model.nint, 
                        self.model.npint), 'D')
        rvy = np.zeros((self.nparam, aq.naq, self.model.nint, 
                        self.model.npint), 'D')
        if aq == self.aq:
            qxqy = np.zeros((2 * (self.order + 1), self.model.npint), 'D')
            for i in range(self.aq.naq):
                for j in range(self.model.nint):
                    if besselnumba.isinside(self.z1, self.z2, x + y * 1j, 
                                       self.rzero * self.aq.lababs[i, j]):
                        qxqy[:, :] = besselnumba.bessellsqxqyv2(x, y, 
                            self.z1, self.z2, self.aq.lab2[i, j, :], 
                            self.order, 
                            self.rzero * self.aq.lababs[i, j]) / self.L
                    for k in range(self.nlayers):
                        rvx[k::self.nlayers, i, j, :] = \
                            self.term2[k, i, j, :] * qxqy[:self.order + 1, :]
                        rvy[k::self.nlayers, i, j, :] = \
                            self.term2[k, i, j, :] * qxqy[self.order + 1:, :]
        rvx.shape = (self.nparam, aq.naq, self.model.npval)
        rvy.shape = (self.nparam, aq.naq, self.model.npval)
        return rvx, rvy

    def headinside(self, t):
        """The head inside the line-sink
        
        Returns
        -------
        array (length number of screens)
            Head inside the well for each screen
            
        """
        
        return self.model.head(self.xc, self.yc, t)[self.layers] - \
               self.resfach[:, np.newaxis] * self.discharge(t)

    def plot(self):
        plt.plot([self.x1, self.x2], [self.y1, self.y2], 'k')
    
class HeadLineSinkHo(LineSinkHoBase, HeadEquationNores):
    '''HeadLineSink of which the head varies through time.
    May be screened in multiple layers but all with the same head'''
    
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, tsandh=[(0.0,1.0)],\
                 order=0, layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        if tsandh == 'fixed':
            tsandh = [(0, 0)]
            etype = 'z'
        else:
            etype = 'v'
        LineSinkHoBase.__init__(self, model, x1=x1, y1=y1, x2=x2, y2=y2, 
                                tsandbc=tsandh,  res=0.0, wh='H', order=order, 
                                layers=layers, type=etype,  
                                name='HeadLineSinkHo', label=label, 
                                addtomodel=addtomodel)
        self.nunknowns = self.nparam

    def initialize(self):
        LineSinkHoBase.initialize(self)
        self.parameters = np.zeros((self.model.ngvbc, self.nparam, 
                                    self.model.npval), 'D')
        self.pc = np.empty(self.nparam)
        for i, T in enumerate(self.aq.T[self.layers]):
            # Needed in solving; solve for a unit head
            self.pc[i::self.nlayers] =  T 


# Copyright 2019 Kristopher L. Kuhlman <klkuhlm _at_ sandia _dot_ gov>

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

class InverseLaplaceTransform(object):

    def __init__(self):
        self.talbot_cache = {}
        self.stehfest_cache = {}

    def clear(self):
        self.talbot_cache = {}
        self.stehfest_cache = {}

    def calc_laplace_parameter(self,t,**kwargs):
        raise NotImplementedError

    def calc_time_domain_solution(self,fp):
        raise NotImplementedError

class FixedTalbot(InverseLaplaceTransform):
    def calc_laplace_parameter(self,t,**kwargs):

        import numpy as np
        
        # required
        # ------------------------------
        # time of desired approximation
        self.t = t

        # optional
        # ------------------------------
        # maximum time desired (used for scaling) default is requested
        # time.
        self.tmax = kwargs.get('tmax',self.t)

        self.degree = int(kwargs.get('degree',14))
        M = self.degree

        # Abate & Valko rule of thumb for r parameter
        if 'r' in kwargs:
            self.r = kwargs['r']
            r_default = False
        else:
            self.r = 2.0/5.0*M
            r_default = True

        if r_default and self.degree in self.talbot_cache:
            self.theta,self.cot_theta,self.delta = self.talbot_cache[self.degree]
        else:
            self.theta = np.linspace(0.0, np.pi, M+1, dtype=np.float64)
            self.cot_theta = np.empty((M,), dtype=np.float64)

            self.cot_theta[0] = 0.0
            self.cot_theta[1:] = 1.0/np.tan(self.theta[1:-1])

            # all but time-dependent part of p
            self.delta = np.empty((M,), dtype=np.complex64)
            self.delta[0] = self.r
            self.delta[1:] = self.r*self.theta[1:-1]*(self.cot_theta[1:] + 1j)

            self.talbot_cache[self.degree] = self.theta,self.cot_theta,self.delta
            
        self.p = self.delta/self.tmax

        # NB: p is complex

    def calc_time_domain_solution(self,fp,t):

        import numpy as np
        
        # required
        # ------------------------------
        self.t = t

        # assume fp was computed from p matrix returned from
        # calc_laplace_parameter()

        # these were computed in previous call to
        # calc_laplace_parameter()
        theta = self.theta
        delta = self.delta
        M = self.degree
        p = self.p
        r = self.r

        ans = np.empty((M,), dtype=np.complex64)
        ans[0] = np.exp(delta[0])*fp[0]/2.0

        ans[1:] = np.exp(delta[1:])*fp[1:]
        ans[1:] *= (1.0 + 1j*theta[1:-1]*(1.0 + self.cot_theta[1:]**2) -
                    1j*self.cot_theta[1:])

        result = 2.0/5.0*np.sum(ans)/self.t

        # ignore any small imaginary part
        return result.real

# ****************************************

class Stehfest(InverseLaplaceTransform):

    def calc_laplace_parameter(self,t,**kwargs):

        import numpy as np
        
        # required
        # ------------------------------
        # time of desired approximation
        self.t = t

        # optional
        # ------------------------------

        self.degree = int(kwargs.get('degree',16))

        self.ln2 = np.log(2.0)
        
        # _coeff routine requires even degree
        if self.degree%2 > 0:
            self.degree += 1

        M = self.degree

        if self.degree in self.stehfest_cache:
            self.V = self.stehfest_cache[self.degree]
        else:
            self.V = kwargs.get('V',self._coeff())
            self.stehfest_cache[self.degree] = self.V
            
        self.p = np.arange(1,M+1)*self.ln2/self.t
        
        # NB: p is real

    def _coeff(self):
        r"""Salzer summation weights (aka, "Stehfest coefficients")
        only depend on the approximation order (M) and the precision"""

        import numpy as np
        from scipy.misc import factorial
        
        M = self.degree
        M2 = int(M/2.0) # checked earlier that M is even

        V = np.empty((M,), dtype=np.float64)

        fac = lambda x: float(factorial(x,exact=True))
        
        # Salzer summation weights
        # get very large in magnitude and oscillate in sign,
        # if the precision is not high enough, there will be
        # catastrophic cancellation
        for k in range(1,M+1):
            z = np.zeros((min(k,M2)+1,), dtype=np.float64)
            for j in range(int((k+1)/2.0),min(k,M2)+1):
                z[j] = (j**M2*fac(2*j)/
                        (fac(M2-j)*fac(j)*fac(j-1)*fac(k-j)*fac(2*j-k)))
            V[k-1] = (-1)**(k+M2)*np.sum(z)

        return V

    def calc_time_domain_solution(self,fp,t):

        import numpy as np
        
        # required
        self.t = t

        # assume fp was computed from p matrix returned from
        # calc_laplace_parameter()

        result = np.dot(self.V,fp)*self.ln2/self.t

        # ignore any small imaginary part
        return result.real

# ****************************************

class deHoog(InverseLaplaceTransform):

    def calc_laplace_parameter(self,t,**kwargs):

        import numpy as np

        # NB: too many parameters, and simple p
        # nothing is cached here.
        
        self.t = t
        self.tmax = kwargs.get('tmax',self.t)

        self.degree = int(kwargs.get('degree',17))

        # 2*M+1 terms in approximation
        M = self.degree

        self.alpha = kwargs.get('alpha',1.0E-16)

        # desired tolerance (here simply related to alpha)
        self.tol = kwargs.get('tol',self.alpha*10.0)
        self.nump = 2*self.degree+1 # number of terms in approximation

        # scaling factor (likely tune-able, but 2 is typical)
        self.scale = kwargs.get('scale',2.0)
        self.T = kwargs.get('T',self.scale*self.tmax)

        self.gamma = self.alpha - np.log(self.tol)/(self.scale*self.T)
        self.p = self.gamma + 1j*np.pi*np.arange(self.nump)/self.T
        
        # NB: p is complex (mpc)

    def calc_time_domain_solution(self,fp,t):

        import numpy as np
        
        M = self.degree
        NP = self.nump
        T = self.T

        self.t = t

        # would it be useful to try re-using
        # space between e&q and A&B?
        e = np.empty((NP,M+1), dtype=np.complex64)
        q = np.empty((NP,M), dtype=np.complex64)
        d = np.empty((NP,), dtype=np.complex64)
        A = np.empty((NP+2,), dtype=np.complex64)
        B = np.empty((NP+2,), dtype=np.complex64)

        # initialize Q-D table
        e[0:2*M,0] = 0.0
        q[0,0] = fp[1]/(fp[0]/2.0)
        for i in range(1,2*M):
            q[i,0] = fp[i+1]/fp[i]

        # rhombus rule for filling triangular Q-D table (e & q)
        for r in range(1,M+1):
            # start with e, column 1, 0:2*M-2
            mr = 2*(M-r)
            e[0:mr,r] = q[1:mr+1,r-1] - q[0:mr,r-1] + e[1:mr+1,r-1]
            if not r == M:
                rq = r+1
                mr = 2*(M-rq)+1
                for i in range(mr):
                    q[i,rq-1] = q[i+1,rq-2]*e[i+1,rq-1]/e[i,rq-1]

        # build up continued fraction coefficients (d)
        d[0] = fp[0]/2.0
        for r in range(1,M+1):
            d[2*r-1] = -q[0,r-1] # even terms
            d[2*r]   = -e[0,r]   # odd terms

        # seed A and B for recurrence
        A[0] = 0.0 
        A[1] = d[0]
        B[0:2] = 1.0 

        # base of the power series
        z = np.exp(1j*np.pi*self.t/T) 

        # coefficients of Pade approximation (A & B)
        # using recurrence for all but last term
        for i in range(1,2*M):
            A[i+1] = A[i] + d[i]*A[i-1]*z
            B[i+1] = B[i] + d[i]*B[i-1]*z

        # "improved remainder" to continued fraction
        brem  = (1.0 + (d[2*M-1] - d[2*M])*z)/2.0
        rem = -brem*(1.0 - np.sqrt(1.0 + d[2*M]*z/brem**2))

        # last term of recurrence using new remainder
        A[NP] = A[2*M] + rem*A[2*M-1]
        B[NP] = B[2*M] + rem*B[2*M-1]

        # diagonal Pade approximation
        # F=A/B represents accelerated trapezoid rule
        result = np.exp(self.gamma*self.t)/T*(A[NP]/B[NP]).real

        return result

# ****************************************

# initialize classes
_fixed_talbot = FixedTalbot()
_stehfest = Stehfest()
_de_hoog = deHoog()

def invertlaplace(f, t, **kwargs):

    rule = kwargs.get('method','dehoog')
    if type(rule) is str:
        lrule = rule.lower()
        if lrule == 'talbot':
            rule = _fixed_talbot
        elif lrule == 'stehfest':
            rule = _stehfest
        elif lrule == 'dehoog':
            rule = _de_hoog
        else:
            raise ValueError("unknown invlap algorithm: %s" % rule)
    else:
        rule = rule()

    # determine the vector of Laplace-space parameter
    # needed for the requested method and desired time
    rule.calc_laplace_parameter(t,**kwargs)

    # compute the Laplace-space function evalutations
    # at the required abscissa.
    fp = [f(p) for p in rule.p]

    # compute the time-domain solution from the
    # Laplace-space function evaluations
    return rule.calc_time_domain_solution(fp,t)

# shortcuts for the above function for specific methods
def invlaptalbot(*args, **kwargs):
    kwargs['method'] = 'talbot'
    return invertlaplace(*args, **kwargs)

def invlapstehfest(*args, **kwargs):
    kwargs['method'] = 'stehfest'
    return invertlaplace(*args, **kwargs)

def invlapdehoog(*args, **kwargs):
    kwargs['method'] = 'dehoog'
    return invertlaplace(*args, **kwargs)




'''
Copyright (C), 2017, Mark Bakker.
Mark Bakker, Delft University of Technology
mark dot bakker at tudelft dot nl

TTim is a computer program for the simulation of transient
multi-layer flow with analytic elements and consists of a
library of Python scripts and FORTRAN extensions.
'''

#--version number
__name__='ttim'
__author__='Mark Bakker'
from .version import __version__

# Import all classes and functions
from .model import ModelMaq, Model3D
from .well import DischargeWell, HeadWell, Well, WellTest
from .linesink import LineSink, HeadLineSink, HeadLineSinkString, \
     LineSinkDitchString, HeadLineSinkHo
from .linedoublet import LeakyLineDoublet, LeakyLineDoubletString
from .circareasink import CircAreaSink
from .fit import Calibrate
from .trace import timtraceline, timtrace


import numpy as np
import matplotlib.pyplot as plt
import inspect # Used for storing the input

class AquiferData:
    def __init__(self, model, kaq, z, Haq, Hll, c, Saq, Sll, poraq, porll, 
                 ltype, topboundary, phreatictop, kzoverkh=None, model3d=False):
        '''kzoverkh and model3d only need to be specified when model
        is model3d'''
        self.model = model
        self.kaq = np.atleast_1d(kaq).astype('d')
        self.z = np.atleast_1d(z).astype('d')
        self.naq = len(self.kaq)
        self.nlayers = len(self.z) - 1
        self.Haq = np.atleast_1d(Haq).astype('d')
        self.Hll = np.atleast_1d(Hll).astype('d')
        self.T = self.kaq * self.Haq
        self.Tcol = self.T.reshape(self.naq, 1)
        self.c = np.atleast_1d(c).astype('d')
        self.c[self.c > 1e100] = 1e100 
        self.Saq = np.atleast_1d(Saq).astype('d')
        self.Sll = np.atleast_1d(Sll).astype('d')
        self.Sll[self.Sll < 1e-20] = 1e-20 # Cannot be zero
        self.poraq = np.atleast_1d(poraq).astype('d')
        self.porll = np.atleast_1d(porll).astype('d')
        self.ltype = np.atleast_1d(ltype)
        self.zaqtop = self.z[:-1][self.ltype == 'a']
        self.zaqbot = self.z[1:][self.ltype == 'a']
        self.layernumber = np.zeros(self.nlayers, dtype='int')
        self.layernumber[self.ltype == 'a'] = np.arange(self.naq)
        self.layernumber[self.ltype == 'l'] = np.arange(self.nlayers - self.naq)
        if self.ltype[0] == 'a':
            self.layernumber[self.ltype == 'l'] += 1  # first leaky layer below first aquifer layer
        self.topboundary = topboundary[:3]
        self.phreatictop = phreatictop
        self.kzoverkh = kzoverkh
        if self.kzoverkh is not None:
            self.kzoverkh = np.atleast_1d(self.kzoverkh).astype('d')
            if len(self.kzoverkh) == 1: 
                self.kzoverkh = self.kzoverkh * np.ones(self.naq)
        self.model3d = model3d
        if self.model3d:
            assert self.kzoverkh is not None, \
                "model3d specified without kzoverkh"
        #self.D = self.T / self.Saq
        self.area = 1e200 # Smaller than default of ml.aq so that inhom is found
    
    def __repr__(self):
        return 'Inhom T: ' + str(self.T)
    
    def initialize(self):
        '''
        eigval[naq, npval]: Array with eigenvalues
        lab[naq, npval]: Array with lambda values
        lab2[naq, nint, npint]: Array with lambda values reorganized per
        interval
        eigvec[naq, naq, npval]: Array with eigenvector matrices
        coef[naq ,naq, npval]: Array with coefficients;
        coef[ilayers, :, np] are the coefficients if the element is in
        ilayers belonging to Laplace parameter number np
        '''
        # Recompute T for when kaq is changed
        self.T = self.kaq * self.Haq
        self.Tcol = self.T.reshape(self.naq, 1)
        # Compute Saq and Sll
        self.Scoefaq = self.Saq * self.Haq
        self.Scoefll = self.Sll * self.Hll
        if (self.topboundary == 'con') and self.phreatictop:
            self.Scoefaq[0] = self.Scoefaq[0] / self.Haq[0]
        elif (self.topboundary == 'lea') and self.phreatictop:
            self.Scoefll[0] = self.Scoefll[0] / self.Hll[0]
        self.D = self.T / self.Scoefaq
        # Compute c if model3d for when kaq is changed
        if self.model3d:
            self.c[1:] = \
                0.5 * self.Haq[:-1] / (self.kzoverkh[:-1] * self.kaq[:-1]) + \
                0.5 * self.Haq[1:] /  (self.kzoverkh[1:] * self.kaq[1:])
        #
        self.eigval = np.zeros((self.naq, self.model.npval), 'D')
        self.lab = np.zeros((self.naq, self.model.npval), 'D')
        self.eigvec = np.zeros((self.naq, self.naq, self.model.npval), 'D')
        self.coef = np.zeros((self.naq, self.naq, self.model.npval), 'D')
        b = np.diag(np.ones(self.naq))
        for i in range(self.model.npval):
            w, v = self.compute_lab_eigvec(self.model.p[i]) 
            # Eigenvectors are columns of v
            self.eigval[:, i] = w; self.eigvec[:, :, i] = v
            self.coef[:, :, i] = np.linalg.solve(v, b).T
        self.lab = 1.0 / np.sqrt(self.eigval)
        self.lab2 = self.lab.copy() 
        self.lab2.shape = (self.naq, self.model.nint, self.model.npint)
        self.lababs = np.abs(self.lab2[:, :, 0]) # used to check distances
        self.eigvec2 = self.eigvec.copy()
        self.eigvec2.shape = (self.naq, self.naq, 
                              self.model.nint, self.model.npint)
    
    def compute_lab_eigvec(self, p, returnA = False, B = None):
        sqrtpSc = np.sqrt( p * self.Scoefll * self.c )
        a, b = np.zeros_like(sqrtpSc), np.zeros_like(sqrtpSc)
        small = np.abs(sqrtpSc) < 200
        a[small] = sqrtpSc[small] / np.tanh(sqrtpSc[small])
        b[small] = sqrtpSc[small] / np.sinh(sqrtpSc[small])
        a[~small] = sqrtpSc[~small] / ((1.0 - np.exp(-2.0*sqrtpSc[~small])) / 
                                       (1.0 + np.exp(-2.0*sqrtpSc[~small])))
        b[~small] = sqrtpSc[~small] * 2.0 * np.exp(-sqrtpSc[~small]) / \
                    (1.0 - np.exp(-2.0*sqrtpSc[~small]))
        if (self.topboundary[:3] == 'sem') or (self.topboundary[:3] == 'lea'):
            dzero = sqrtpSc[0] * np.tanh(sqrtpSc[0])

        d0 = p / self.D
        if B is not None:
            d0 = d0 * B  # B is vector of load efficiency paramters
        d0[:-1] += a[1:] / (self.c[1:] * self.T[:-1])
        d0[1:]  += a[1:] / (self.c[1:] * self.T[1:])
        if self.topboundary[:3] == 'lea':
            d0[0] += dzero / ( self.c[0] * self.T[0] )
        elif self.topboundary[:3] == 'sem':
            d0[0] += a[0] / ( self.c[0] * self.T[0] )
            
        dm1 = -b[1:] / (self.c[1:] * self.T[:-1])
        dp1 = -b[1:] / (self.c[1:] * self.T[1:])
        A = np.diag(dm1,-1) + np.diag(d0,0) + np.diag(dp1,1)
        if returnA: return A
        w, v = np.linalg.eig(A)
        # sorting moved here
        index = np.argsort(abs(w))[::-1]
        w = w[index]
        v = v[:, index]
        return w, v
    
    def head_to_potential(self, h, layers):
        return h * self.Tcol[layers]
    
    def potential_to_head(self, pot, layers):
        return pot / self.Tcol[layers]
    
    def isInside(self,x,y):
        print('Must overload AquiferData.isInside method')
        return True
    
    def inWhichLayer(self, z):
        '''Returns -9999 if above top of system, 
        +9999 if below bottom of system, 
        negative for in leaky layer.
        leaky layer -n is on top of aquifer n'''
        if z > self.zt[0]:
            return -9999
        for i in range(self.naq-1):
            if z >= self.zb[i]:
                return i
            if z > self.zt[i+1]:
                return -i-1
        if z >= self.zb[self.naq-1]:
            return self.naq - 1
        return +9999
    
    def findlayer(self, z):
        '''
        Returns layer-number, layer-type and model-layer-number'''
        if z > self.z[0]:
            modellayer = -1
            ltype = 'above'
            layernumber = None
        elif z < self.z[-1]:
            modellayer = len(self.layernumber)
            ltype = 'below'
            layernumber = None
        else:
            modellayer = np.argwhere((z <= self.z[:-1]) & 
                                     (z >= self.z[1:]))[0, 0]
            layernumber = self.layernumber[modellayer]
            ltype = self.ltype[modellayer] 
        return layernumber, ltype, modellayer
    
class Aquifer(AquiferData):
    def __init__(self, model, kaq, z, Haq, Hll, c, Saq, Sll, poraq, porll, 
                 ltype, topboundary, phreatictop, kzoverkh=None, model3d=False):
        AquiferData.__init__(self, model, kaq, z, Haq, Hll, c, Saq, Sll, 
            poraq, porll, ltype, topboundary, phreatictop, kzoverkh, model3d)
        self.inhomlist = []
        self.area = 1e300 # Needed to find smallest inhomogeneity
    
    def __repr__(self):
        return 'Background Aquifer T: ' + str(self.T)
    
    
    def initialize(self):
        AquiferData.initialize(self)
        for inhom in self.inhomlist:
            inhom.initialize()
    
    def find_aquifer_data(self, x, y):
        rv = self
        for aq in self.inhomlist:
            if aq.isInside(x, y):
                if aq.area < rv.area:
                    rv = aq
        return rv


import numpy as np

class HeadEquation:
    def equation(self):
        '''Mix-in class that returns matrix rows for head-specified conditions. 
        (really written as constant potential element)
        Works for nunknowns = 1
        Returns matrix part nunknowns,neq,npval, complex
        Returns rhs part nunknowns,nvbc,npval, complex
        Phi_out - c*T*q_s = Phi_in
        Well: q_s = Q / (2*pi*r_w*H)
        LineSink: q_s = sigma / H = Q / (L*H)
        '''
        mat = np.empty((self.nunknowns, self.model.neq, 
                        self.model.npval), 'D')
        # rhs needs be initialized zero
        rhs = np.zeros((self.nunknowns, self.model.ngvbc, 
                        self.model.npval), 'D')
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0  
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    mat[istart: istart + self.nlayers, 
                        ieq: ieq + e.nunknowns, :] = e.potinflayers(
                            self.xc[icp], self.yc[icp], self.layers)
                    if e == self:
                        for i in range(self.nlayers): 
                            mat[istart + i, ieq + istart + i, :] -= \
                                self.resfacp[istart + i] * \
                                e.dischargeinflayers[istart + i]
                    ieq += e.nunknowns
            for i in range(self.model.ngbc):
                rhs[istart: istart + self.nlayers, i, :] -= \
                self.model.gbclist[i].unitpotentiallayers(
                    self.xc[icp], self.yc[icp], self.layers)
            if self.type == 'v':
                iself = self.model.vbclist.index(self)
                for i in range(self.nlayers):
                    rhs[istart + i, self.model.ngbc + iself, :] = \
                        self.pc[istart + i] / self.model.p
        return mat, rhs
    
class WellBoreStorageEquation:
    def equation(self):
        '''Mix-in class that returns matrix rows for multi-aquifer element with
        total given discharge, uniform but unknown head and 
        InternalStorageEquation
        '''
        mat = np.zeros((self.nunknowns, self.model.neq, 
                        self.model.npval), 'D')
        rhs = np.zeros((self.nunknowns, self.model.ngvbc,
                        self.model.npval), 'D')
        ieq = 0
        for e in self.model.elementlist:
            if e.nunknowns > 0:
                head = e.potinflayers(self.xc[0], self.yc[0], self.layers) / \
                       self.aq.T[self.layers][:, np.newaxis, np.newaxis]
                mat[:-1, ieq: ieq + e.nunknowns, :] = head[:-1, :] - head[1:, :]
                mat[-1, ieq: ieq + e.nunknowns, :] -= np.pi * self.rc**2 * \
                                                      self.model.p * head[0, :]
                if e == self:
                    disterm = self.dischargeinflayers * self.res / (2 * np.pi * 
                        self.rw * self.aq.Haq[self.layers][:, np.newaxis])
                    if self.nunknowns > 1:  # Multiple layers
                        for i in range(self.nunknowns - 1):
                            mat[i, ieq + i, :] -= disterm[i]
                            mat[i, ieq + i + 1, :] += disterm[i + 1]
                    mat[-1, ieq: ieq + self.nunknowns, :] += \
                        self.dischargeinflayers
                    mat[-1, ieq, :] += \
                        np.pi * self.rc ** 2 * self.model.p * disterm[0]
                ieq += e.nunknowns
        for i in range(self.model.ngbc):
            head = self.model.gbclist[i].unitpotentiallayers(
                self.xc[0], self.yc[0], self.layers) / \
                self.aq.T[self.layers][:, np.newaxis]
            rhs[:-1, i, :] -= head[:-1, :] - head[1:, :]
            rhs[-1, i, :] += np.pi * self.rc ** 2 * self.model.p * head[0, :]
        if self.type == 'v':
            iself = self.model.vbclist.index(self)
            rhs[-1, self.model.ngbc + iself, :] += self.flowcoef
            if self.hdiff is not None:
                # head[0] - head[1] = hdiff
                rhs[:-1, self.model.ngbc + iself, :] += \
                    self.hdiff[:, np.newaxis] / self.model.p  
        return mat, rhs
    
class HeadEquationNores:
    def equation(self):
        '''Mix-in class that returns matrix rows for head-specified conditions. 
        (really written as constant potential element)
        Returns matrix part nunknowns, neq, npval, complex
        Returns rhs part nunknowns, nvbc, npval, complex
        '''
        mat = np.empty((self.nunknowns, self.model.neq, 
                        self.model.npval), 'D')
        rhs = np.zeros((self.nunknowns, self.model.ngvbc, 
                        self.model.npval), 'D')
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0  
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    mat[istart: istart + self.nlayers, 
                        ieq: ieq + e.nunknowns, :] = e.potinflayers(
                        self.xc[icp], self.yc[icp], self.layers)
                    ieq += e.nunknowns
            for i in range(self.model.ngbc):
                rhs[istart: istart + self.nlayers, i, :] -= \
                    self.model.gbclist[i].unitpotentiallayers(
                    self.xc[icp], self.yc[icp], self.layers)
            if self.type == 'v':
                iself = self.model.vbclist.index(self)
                for i in range(self.nlayers):
                    rhs[istart + i, self.model.ngbc + iself, :] = \
                        self.pc[istart + i] / self.model.p
        return mat, rhs
    
class LeakyWallEquation:
    def equation(self):
        '''Mix-in class that returns matrix rows for leaky-wall condition
        Returns matrix part nunknowns,neq,npval, complex
        Returns rhs part nunknowns,nvbc,npval, complex
        '''
        mat = np.empty((self.nunknowns, self.model.neq,
                        self.model.npval), 'D')
        rhs = np.zeros((self.nunknowns, self.model.ngvbc,
                        self.model.npval), 'D')
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0  
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    qx, qy = e.disvecinflayers(self.xc[icp], self.yc[icp], 
                                               self.layers)
                    mat[istart: istart + self.nlayers, 
                        ieq: ieq + e.nunknowns, :] = \
                        qx * self.cosout[icp] + qy * self.sinout[icp]
                    if e == self:
                        hmin = e.potinflayers(
                            self.xcneg[icp], self.ycneg[icp], self.layers) / \
                            self.aq.T[self.layers][: ,np.newaxis, np.newaxis]
                        hplus = e.potinflayers(
                            self.xc[icp], self.yc[icp], self.layers) / \
                            self.aq.T[self.layers][:, np.newaxis, np.newaxis]
                        mat[istart:istart + self.nlayers, 
                            ieq: ieq + e.nunknowns, :] -= \
                            self.resfac[:, np.newaxis, np.newaxis] * \
                            (hplus - hmin)
                    ieq += e.nunknowns
            for i in range(self.model.ngbc):
                qx, qy = self.model.gbclist[i].unitdisveclayers(
                    self.xc[icp], self.yc[icp], self.layers)
                rhs[istart: istart + self.nlayers, i, :] -=  \
                    qx * self.cosout[icp] + qy * self.sinout[icp]
            #if self.type == 'v':
            #    iself = self.model.vbclist.index(self)
            #    for i in range(self.nlayers):
            #        rhs[istart+i,self.model.ngbc+iself,:] = \
            #             self.pc[istart+i] / self.model.p
        return mat, rhs
    
class MscreenEquation:
    def equation(self):
        '''Mix-in class that returns matrix rows for multi-screen conditions 
        where total discharge is specified.
        Works for nunknowns = 1
        Returns matrix part nunknowns, neq, npval, complex
        Returns rhs part nunknowns, nvbc, npval, complex
        head_out - c * q_s = h_in
        Set h_i - h_(i + 1) = 0 and Sum Q_i = Q'''
        mat = np.zeros((self.nunknowns, self.model.neq, 
                        self.model.npval), 'D')
        rhs = np.zeros((self.nunknowns, self.model.ngvbc, 
                        self.model.npval), 'D')
        ieq = 0
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0 
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    head = e.potinflayers(
                        self.xc[icp], self.yc[icp], self.layers) / \
                        self.aq.T[self.layers][:,np.newaxis,np.newaxis]
                    mat[istart: istart + self.nlayers - 1, 
                        ieq: ieq + e.nunknowns, :] = \
                        head[:-1,:] - head[1:,:]
                    if e == self:
                        for i in range(self.nlayers-1):
                            mat[istart + i, ieq + istart + i, :] -= \
                                self.resfach[istart + i] * \
                                e.dischargeinflayers[istart + i]
                            mat[istart + i, ieq + istart + i + 1, :] += \
                                self.resfach[istart + i + 1] * \
                                e.dischargeinflayers[istart + i + 1]
                            mat[istart + i, 
                                ieq + istart: ieq + istart + i + 1, :] -= \
                                self.vresfac[istart + i] * \
                                e.dischargeinflayers[istart + i]
                        mat[istart + self.nlayers - 1,
                            ieq + istart: ieq + istart + self.nlayers, :] = 1.0
                    ieq += e.nunknowns
            for i in range(self.model.ngbc):
                head = self.model.gbclist[i].unitpotentiallayers(
                    self.xc[icp], self.yc[icp], self.layers) / \
                    self.aq.T[self.layers][:, np.newaxis]
                rhs[istart: istart + self.nlayers - 1, i, :] -= \
                    head[:-1,:] - head[1:,:]
            if self.type == 'v':
                iself = self.model.vbclist.index(self)
                rhs[istart + self.nlayers - 1, self.model.ngbc + iself, :] = 1.0  
            # If self.type == 'z', it should sum to zero, 
            # which is the default value of rhs
        return mat, rhs
    
class MscreenDitchEquation:
    def equation(self):
        '''Mix-in class that returns matrix rows for multi-screen conditions 
        where total discharge is specified.
        Returns matrix part nunknowns,neq,npval, complex
        Returns rhs part nunknowns,nvbc,npval, complex
        head_out - c*q_s = h_in
        Set h_i - h_(i+1) = 0 and Sum Q_i = Q
        I would say
        headin_i - headin_(i+1) = 0
        headout_i - c*qs_i - headout_(i+1) + c*qs_(i+1) = 0 
        In case of storage:
        Sum Q_i - A * p^2 * headin = Q
        '''
        mat = np.zeros((self.nunknowns, self.model.neq,
                        self.model.npval), 'D')
        rhs = np.zeros((self.nunknowns, self.model.ngvbc,
                        self.model.npval), 'D')
        ieq = 0
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0 
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    head = e.potinflayers(
                        self.xc[icp], self.yc[icp], self.layers) / \
                        self.aq.T[self.layers][:, np.newaxis, np.newaxis]
                    if self.nlayers > 1: 
                        mat[istart: istart + self.nlayers - 1,
                            ieq: ieq + e.nunknowns, :] = \
                            head[:-1, :] - head[1:, :]
                    # Store head in top layer in 2nd to last equation 
                    # of this control point
                    mat[istart + self.nlayers - 1, 
                        ieq: ieq + e.nunknowns, :] = head[0,:] 
                    if e == self:
                        # Correct head in top layer in second to last equation 
                        # to make it head inside
                        mat[istart + self.nlayers - 1,
                            ieq + istart, :] -= self.resfach[istart] * \
                            e.dischargeinflayers[istart]
                        if icp == 0:
                            istartself = ieq  # Needed to build last equation
                        for i in range(self.nlayers-1):
                            mat[istart + i, ieq + istart + i, :] -= \
                                self.resfach[istart + i] * \
                                e.dischargeinflayers[istart + i]
                            mat[istart + i, ieq + istart + i + 1, :] += \
                                self.resfach[istart + i + 1] * \
                                e.dischargeinflayers[istart + i + 1]
                            #vresfac not yet used here; it is set to zero as 
                            #I don't quite now what is means yet
                            #mat[istart + i, ieq + istart:ieq+istart+i+1,:] -= \
                            # self.vresfac[istart + i] * \
                            # e.dischargeinflayers[istart + i]
                    ieq += e.nunknowns
            for i in range(self.model.ngbc):
                head = self.model.gbclist[i].unitpotentiallayers(
                    self.xc[icp], self.yc[icp], self.layers) / \
                    self.aq.T[self.layers][:, np.newaxis]
                if self.nlayers > 1: 
                    rhs[istart: istart + self.nlayers - 1, i, :] -= \
                        head[:-1, :] - head[1:, :]
                # Store minus the head in top layer in second to last equation 
                # for this control point
                rhs[istart + self.nlayers - 1, i, :] -= head[0, :] 
        # Modify last equations
        for icp in range(self.ncp - 1):
            ieq = (icp + 1) * self.nlayers - 1
            # Head first layer control point icp - Head first layer control
            # point icp + 1
            mat[ieq, :, :] -= mat[ieq + self.nlayers, :, :]  
            rhs[ieq, :, :] -= rhs[ieq + self.nlayers, :, :]
        # Last equation setting the total discharge of the ditch
        mat[-1, :, :] = 0.0  
        mat[-1, istartself: istartself + self.nparam, :] = 1.0
        if self.Astorage is not None:
            # Used to store last equation in case of ditch storage
            matlast = np.zeros((self.model.neq, self.model.npval), 'D')
            rhslast = np.zeros((self.model.npval), 'D')  
            ieq = 0
            for e in self.model.elementlist:
                head = e.potinflayers(self.xc[0], self.yc[0], self.layers) / \
                    self.aq.T[self.layers][:, np.newaxis, np.newaxis]
                matlast[ieq: ieq + e.nunknowns] -= \
                    self.Astorage * self.model.p ** 2 * head[0, :]
                if e == self:
                    # only need to correct first unknown 
                    matlast[ieq] += self.Astorage * self.model.p ** 2 * \
                        self.resfach[0] * e.dischargeinflayers[0]
                ieq += e.nunknowns
            for i in range(self.model.ngbc):
                head = self.model.gbclist[i].unitpotentiallayers(
                    self.xc[0], self.yc[0], self.layers) / \
                    self.aq.T[self.layers][:, np.newaxis]
                rhslast += self.Astorage * self.model.p ** 2 * head[0] 
            mat[-1] += matlast
        rhs[-1, :, :] = 0.0
        if self.type == 'v':
            iself = self.model.vbclist.index(self)
            rhs[-1, self.model.ngbc + iself, :] = 1.0  
            # If self.type == 'z', it should sum to zero, which is the default 
            # value of rhs
            if self.Astorage is not None: 
                rhs[-1, self.model.ngbc + iself, :] += rhslast
        return mat, rhs
    
class InhomEquation:
    def equation(self):
        '''Mix-in class that returns matrix rows for inhomogeneity conditions'''
        mat = np.zeros((self.nunknowns, self.model.neq,
                        self.model.npval), 'D')
        rhs = np.zeros((self.nunknowns, self.model.ngvbc,
                        self.model.npval), 'D')
        for icp in range(self.ncp):
            istart = icp * 2 * self.nlayers
            ieq = 0  
            for e in self.model.elementList:
                if e.nunknowns > 0:
                    mat[istart: istart + self.nlayers,
                        ieq: ieq + e.nunknowns, :] = \
                    e.potinflayers(self.xc[icp], self.yc[icp], 
                                   self.layers, self.aqin) / \
                    self.aqin.T[self.layers][:, np.newaxis, np.newaxis] - \
                    e.potinflayers(self.xc[icp], self.yc[icp], 
                                   self.layers, self.aqout) / \
                    self.aqout.T[self.layers][:, np.newaxis, np.newaxis]
                    qxin, qyin = e.disinflayers(
                        self.xc[icp], self.yc[icp], self.layers, self.aqin)
                    qxout, qyout = e.disinflayers(
                        self.xc[icp], self.yc[icp], self.layers, self.aqout)
                    mat[istart + self.nlayers: istart + 2 * self.nlayers,
                        ieq: ieq + e.nunknowns, :] = \
                        (qxin - qxout) * np.cos(self.thetacp[icp]) + \
                        (qyin - qyout) * np.sin(self.thetacp[icp])
                    ieq += e.nunknowns
            for i in range(self.model.ngbc):
                rhs[istart: istart + self.nlayers, i, :] -= (
                    self.model.gbclist[i].unitpotentiallayers(
                    self.xc[icp], self.yc[icp], self.layers, self.aqin) /
                    self.aqin.T[self.layers][:, np.newaxis] - 
                    self.model.gbclist[i].unitpotentiallayers(
                    self.xc[icp], self.yc[icp], self.layers, self.aqout) /
                    self.aqout.T[self.layers][:, np.newaxis])
                qxin, qyin = self.model.gbclist[i].unitdischargelayers(
                    self.xc[icp], self.yc[icp], self.layers, self.aqin)
                qxout,qyout = self.model.gbclist[i].unitdischargelayers(
                    self.xc[icp], self.yc[icp], self.layers, self.aqout)
                rhs[istart + self.nlayers: istart + 2 * self.nlayers, i, :] -= \
                    (qxin - qxout) * np.cos(self.thetacp[icp]) + \
                    (qyin - qyout) * np.sin(self.thetacp[icp])
        return mat, rhs


import numpy as np
import matplotlib.pyplot as plt

class PlotTtim:
    def plot(self, win=None, newfig=True, figsize=None):
        """Plot layout
        
        Parameters
        ----------
    
        win : list or tuple
            [x1, x2, y1, y2]
            
        """
        
        if newfig:
            plt.figure(figsize=figsize)
            ax1 = plt.subplot()
        else:
            fig = plt.gcf()
            ax1 = fig.axes[0]
        if ax1 is not None:
            plt.sca(ax1)
            for e in self.elementlist:
                e.plot()
            plt.axis('scaled')
            if win is not None:
                plt.axis(win)
    
    def xsection(self, x1=0, x2=1, y1=0, y2=0, npoints=100, t=1, layers=0,
                 sstart=0, color=None, lw=1, figsize=None, newfig=True,
                 legend=True):
        layers = np.atleast_1d(layers)
        if newfig:
            plt.figure(figsize=figsize)
        x = np.linspace(x1, x2, npoints)
        y = np.linspace(y1, y2, npoints)
        s = np.sqrt((x - x[0]) ** 2 + (y - y[0]) ** 2 ) + sstart
        h = self.headalongline(x, y, t, layers)
        nlayers, ntime, npoints = h.shape
        for i in range(nlayers):
            for j in range(ntime):
                if color is None:
                    plt.plot(s, h[i, j, :], lw=lw)
                else:
                    plt.plot(s, h[i, j, :], color, lw=lw)
        if legend:
            legendlist = ['layer ' + str(i) for i in layers]
            plt.legend(legendlist)
        #plt.draw()
        
    def contour(self, win, ngr=20, t=1, layers=0, levels=20, layout=True, 
                labels=False, decimals=0, color=None, newfig=True, 
                figsize=None, legend=True):
        """Contour plot
        
        Parameters
        ----------
    
        win : list or tuple
            [x1, x2, y1, y2]
        ngr : scalar, tuple or list
            if scalar: number of grid points in x and y direction
            if tuple or list: nx, ny, number of grid points in x and y direction
        t : scalar
            time
        layers : integer, list or array
            layers for which grid is returned
        levels : integer or array (default 20)
            levels that are contoured
        layout : boolean (default True_)
            plot layout of elements
        labels : boolean (default True)
            print labels along contours
        decimals : integer (default 0)
            number of decimals of labels along contours
        color : str or list of strings
            color of contour lines
        newfig : boolean (default True)
            create new figure
        figsize : tuple of 2 values (default is mpl default)
            size of figure
        legend : list or boolean (default True)
            add legend to figure
            if list of strings: use strings as names in legend
            
        """
        
        x1, x2, y1, y2 = win
        if np.isscalar(ngr):
            nx = ny = ngr
        else:
            nx, ny = ngr
        layers = np.atleast_1d(layers)
        xg = np.linspace(x1, x2, nx)
        yg = np.linspace(y1, y2, ny)
        h = self.headgrid(xg, yg, t, layers)
        if newfig:
            plt.figure(figsize=figsize)
        # color
        if color is None:
            c = plt.rcParams['axes.prop_cycle'].by_key()['color']
        elif type(color) is str:
            c = len(layers) * [color]
        elif type(color) is list:
            c = color
        if len(c) < len(layers):
            n = np.ceil(self.aq.naq / len(c))
            c = n * c 
        # contour
        cscollectionlist = []
        for i in range(len(layers)):
            cs = plt.contour(xg, yg, h[i, 0], levels, colors=c[i], 
                             linestyles='-')
            cscollectionlist.append(cs.collections[0])
            if labels:
                fmt = '%1.' + str(decimals) + 'f'
                plt.clabel(cs, fmt=fmt)
        if type(legend) is list:
            plt.legend(cscollectionlist, legend)
        elif legend:
            legendlist = ['layer ' + str(i) for i in layers]
            plt.legend(cscollectionlist, legendlist)
        plt.axis('scaled')
        if layout:
            self.plot(win=[x1, x2, y1, y2], newfig=False)